#! /usr/bin/env python

from __future__ import with_statement

import os
import platform
import subprocess
import sys
from contextlib import contextmanager


PYTHON_VERSION = ".".join((str(sys.version_info[0]), str(sys.version_info[1])))
OS_VERSION = tuple(int(v) for v in platform.mac_ver()[0].split('.'))


@contextmanager
def workingdirectory(directory):
    wd = os.getcwd()
    os.chdir(directory)
    yield directory
    os.chdir(wd)


class Builder(object):

    if OS_VERSION[1] < 4:
        raise Exception("This version of Mac OS X is not supported because it is too old.")
    elif OS_VERSION[1] == 4:
        CC = 'gcc-4.0'
    elif OS_VERSION[1] == 5:
        CC = 'gcc-4.0'
    elif OS_VERSION[1] == 6:
        CC = 'llvm-gcc-4.2'
    else:
        raise Exception("Unknown version of Mac OS X.")

    def __init__(self, options):
        self.options = options
        self.root = os.path.dirname(os.path.abspath(sys.argv[0]))
        self.dstroot = None
        self.srcroot = os.path.join(self.root, 'bazaar.src')
        self.pkgroot = os.path.join(self.root, 'package')
        if OS_VERSION[1] == 4:
            self.sitepackages_prefix = 'Library/Frameworks/Python.framework/Versions/%s/lib/python%s/site-packages' % \
                (sys.version[:3], sys.version[:3])
        else:
            self.sitepackages_prefix = os.path.join('Library/Python', sys.version[:3], 'site-packages')

# FS utilities

    @contextmanager
    def dstrootdirectory(self, directory):
        old_dstroot = self.dstroot
        self.dstroot = os.path.join(self.root, 'package', os.path.basename(directory) + '.dst')
        if not os.path.isdir(self.dstroot):
            os.mkdir(self.dstroot)
        yield self.dstroot
        self.dstroot = old_dstroot

    def subprocess(self, args, env_update=None, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = dict(os.environ)
        if env_update:
            kwargs['env'].update(env_update)
        if 'CC' not in kwargs['env']:
            kwargs['env']['CC'] = self.CC
        if 'shell' not in kwargs:
            kwargs['shell'] = False
        if kwargs['env']['PATH'].find('/Developer/usr/bin') < 0:
            kwargs['env']['PATH'] += ':/Developer/usr/bin'
        if kwargs['env']['PATH'].find('/usr/texbin') < 0:
            kwargs['env']['PATH'] += ':/usr/texbin'
        try:
            subprocess.check_call(args, **kwargs)
        except subprocess.CalledProcessError, e:
            print 'Command failed with non-zero exit status %d' % e.returncode
            print '    command: ' + str(args)
            print '    env: ' + '\n         '.join('%s: %s' % (k, v) for k, v in kwargs['env'].iteritems())
            sys.exit(1)

    def python(self, args, **kwargs):
        args = [sys.executable] + list(args)
        self.subprocess(args, **kwargs)

    def ditto(self, src, dst):
        src = os.path.abspath(src)
        if os.path.isdir(dst):
            dst = os.path.join(os.path.abspath(dst), os.path.basename(src))
        else:
            dst = os.path.abspath(dst)
        self.subprocess(['ditto', src, dst], shell=False)
        return dst

    def dittodir(self, src, dst):
        src = os.path.abspath(src)
        dst = os.path.abspath(dst)
        self.subprocess(['ditto', src, dst], shell=False)
        return dst

    def dstditto(self, src, dst):
        assert(self.dstroot)
        return self.ditto(src, os.path.normpath(self.dstroot + '/' + dst))

    def rm(self, src):
        # use rm to be able to delete both files and folders
        self.subprocess(['rm', '-rf', os.path.abspath(src)])

# building

    def patch_package(self, src, patches):
        with workingdirectory(src):
            for patch, patch_level in patches:
                self.subprocess(['patch', '-f', '-p', str(patch_level), '-i', patch])

    def build_package(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                self.python(['setup.py', 'install', '--root=' + self.dstroot])

    def build_package_bzr(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                self.build_package(src)
                self.subprocess(['make', 'html-sphinx', 'pdf-sphinx'])
                self.rm(os.path.join(self.dstroot, 'System'))
                self.dstditto('bzr.1', 'usr/local/share/man/man1/bzr.1')
                self.ditto('COPYING.txt', self.pkgroot)
                # docs
                docs_dest_root = os.path.join(self.root, 'Documentation')
                self.rm(docs_dest_root)
                os.mkdir(docs_dest_root)
                for doc_prefix in config.docs:
                    docs_src_root = os.path.join(src, doc_prefix)
                    for doc_name in config.docs[doc_prefix]:
                        src_filename = os.path.join(docs_src_root, doc_name)
                        dest_filename = os.path.join(docs_dest_root, config.docs[doc_prefix][doc_name])
                        if os.path.exists(src_filename):
                            print "copying %s to %s" % (src_filename, dest_filename)
                            if os.path.isdir(src_filename):
                                self.dittodir(src_filename, dest_filename)
                            else:
                                self.ditto(src_filename, dest_filename)
        self.bzrlib_path = os.path.join(self.sitepackages_prefix, 'bzrlib')

    def build_package_sip(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                self.sip_bin_dir = os.path.join(self.dstroot, 'usr/local/bin')
                self.sip_site_dir = os.path.join(self.dstroot, self.sitepackages_prefix)
                if not os.path.exists('Makefile'):
                    config = ['-b', self.sip_bin_dir,
                      '-d', self.sip_site_dir,
                      '-e', os.path.join(self.dstroot, 'usr/local/include/python' + PYTHON_VERSION),
                      '-v', os.path.join(self.dstroot, 'usr/local/share/sip'),
                      '--arch=i386',
                      '--arch=x86_64']
                    self.python(['configure.py'] + config)
                self.subprocess(['make'])
                self.subprocess(['make', 'install'])

    def build_package_pyqt(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                if not os.path.exists('Makefile'):
                    config = [
                        '-q', '/usr/bin/qmake',
                        '-d', os.path.join(self.dstroot, self.sitepackages_prefix),
                        '-b', os.path.join(self.dstroot, 'usr/local/bin'),
                        '-e', 'QtCore',
                        '-e', 'QtGui',
                        '--no-sip-files',
                        '--confirm-license',
                        '--no-designer-plugin',
                        '--use-arch=i386'
                    ]
                    self.python(['configure.py'] + config, env_update={
                        'PATH': self.sip_bin_dir + ':' + os.environ['PATH'],
                        'PYTHONPATH': self.sip_site_dir})
                self.subprocess(['make'])
                self.subprocess(['make', 'install'])

    def build_plugin(self, src, dst_dir):
        with self.dstrootdirectory(src):
            os.makedirs(os.path.join(self.dstroot, self.bzrlib_path))
            plugin_dst = self.dstditto(src, os.path.join(self.bzrlib_path, 'plugins', dst_dir))
            with workingdirectory(plugin_dst):
                self.python(['setup.py', 'build_ext', '--inplace'])
        self.rm(os.path.join(plugin_dst, 'build'))

    def build_plugin_bzrtools(self, src, dst_dir):
        with workingdirectory(src):
            self.rm('.be')
        self.build_plugin(src, dst_dir)

    def build_distribution(self):
        with self.dstrootdirectory('bzr'):
            sys.path.insert(0, os.path.join(self.dstroot, self.sitepackages_prefix))
            bzrlib = __import__('bzrlib', globals(), None, [], 0)
        with workingdirectory(self.root):
            print "building installer package"
            args = ['/usr/local/bin/packagesbuild', '-v', '-F', '.', 'Bazaar.pkgproj']
            self.subprocess(args)
            print "building disk image"
            self.rm('dmg')
            os.mkdir('dmg')
            mpkg_filename = 'Bazaar-%s-%s.mpkg' % (bzrlib.version_string,
                                                   config.DISTRIBUTION_VERSION)
            self.ditto('Bazaar.mpkg', os.path.join('dmg', mpkg_filename))
            self.ditto('Documentation', 'dmg/Documentation')
            dmg_filename = 'Bazaar-%s-OSX-%s.%s-%s.dmg' % (
                bzrlib.version_string,
                # We don't care about the minor version
                OS_VERSION[0], OS_VERSION[1],
                config.DISTRIBUTION_VERSION)
            self.rm(dmg_filename)
            self.subprocess(['hdiutil', 'create', '-srcfolder', 'dmg',
                             '-format', 'UDRO',
                             '-volname', 'Bazaar %s' % bzrlib.version_string,
                             dmg_filename])
            self.subprocess(['hdiutil', 'internet-enable', dmg_filename])

    def clean_dstroots(self):
        with workingdirectory(self.pkgroot):
            for item in os.listdir(self.pkgroot):
                if item.endswith('.dst') and os.path.isdir(item):
                    self.rm(item)

    def build_packages(self):
        packages = config.packages
        for package in packages:
            destdir = package['destdir']
            source = config.extract(os.path.join(self.root, 'src',
                                                 config.get_tarfile(package)),
                                    destdir)
            src = self.ditto(source, self.srcroot)
            if package.has_key('patches'):
                patches = [(os.path.join(self.root, 'patches', destdir, patch),
                            patch_level)
                           for patch, patch_level in package['patches']]
                getattr(self, 'pach_package_' + destdir, self.patch_package)(src, patches)
            getattr(self, 'build_package_' + destdir, self.build_package)(src)
            self.rm(os.path.dirname(source))

    def build_plugins(self):
        plugins = config.plugins
        for plugin in plugins:
            source = config.extract(os.path.join(self.root, 'src',
                                                 config.get_tarfile(plugin)),
                                    plugin['destdir'])
            src = self.ditto(source, self.srcroot)
            getattr(self, 'build_plugin_' + plugin['destdir'],
                    self.build_plugin)(src, plugin['destdir'])
            self.rm(os.path.dirname(source))

    def main(self, args):
        if len(args) == 0:
            args = ['build', 'dist']

        for arg in args:
            if arg == 'clean':
                self.rm(self.srcroot)
                self.clean_dstroots()
            elif arg == 'build':
                if not os.path.exists(self.srcroot):
                    os.mkdir(self.srcroot)
                self.clean_dstroots()
                self.build_packages()
                self.build_plugins()
            elif arg == 'dist':
                self.build_distribution()


if __name__ == '__main__':
    import optparse
    parser = optparse.OptionParser(usage='usage: %prog [options] [step [step...]]',
                                   description='Where <step> is clean, build, or dist. Each step must be given for '
                                   'it to execute since there are no implied dependencies between steps. Default is: '
                                   'build dist')
    parser.add_option('-c', '--config', dest='config', action='store', default='config')
    (options, args) = parser.parse_args()

    try:
        global config
        config = __import__(options.config, globals(), locals(), [], 0)
    except ImportError:
        print "ERROR: failed to import config file"
        sys.exit(1)
    try:
        builder = Builder(options)
        builder.main(args)
    except KeyboardInterrupt:
        sys.exit(1)
