#!/usr/bin/env python
from __future__ import with_statement
import os
import subprocess
import sys
import urllib, urllib2, urlparse
from optparse import OptionParser

from bzrlib import errors
from bzrlib.workingtree import WorkingTree

def check_revno(tree_dir, dep):
    if dep['revno'] == 'tip':
        return True

    wt = None
    try:
        try:
            wt = WorkingTree.open_containing(tree_dir)[0]
            wt.lock_read()
        except (errors.NoWorkingTree, errors.NotLocalUrl):
            print >> sys.stderr, "Error: no working tree at %s" % tree_dir
            return False
        if dep['revno'][:4] == 'tag:':
            tag = dep['revno'][4:]
            revid = wt.branch.tags.lookup_tag(tag)
            dep_revno_t = wt.branch.revision_id_to_dotted_revno(revid)
            dep_revno = ".".join(str(n) for n in dep_revno_t)
        else:
            dep_revno = dep['revno']
        revid = wt.last_revision()
        try:
            revno_t = wt.branch.revision_id_to_dotted_revno(revid)
        except errors.NoSuchRevision:
            print >> sys.stderr, "Error: could not map rev id '%s' to revno" % revid
            return False
        revno = ".".join(str(n) for n in revno_t)
    finally:
        if wt is not None:
            wt.unlock()

    match = revno == dep_revno
    if not match:
        print >> sys.stderr, "Warning: revno mismatch in %s, expected: %r but was: %r" % (dep['name'], dep_revno, revno)
    return match


class ProgressSpinner(object):
    """Un spinner *muy* simple para mostrar actividad en la terminal"""

    def __init__(self, out=sys.stderr):
        """inicializa el spinner"""
        self.out = out
        self.running = True
        def spinner_generator():
            while self.running:
                for i in ('|', '/', '-', '\\', '|', '/', '-', '\\'):
                    yield i
        self.spinner = spinner_generator()

    def next(self, *args):
        self.out.write('\r%s' % self.spinner.next())

    def finished(self):
        self.running = False
        self.out.write('\r' + ' '*70 + '\r')


def download(url, destdir, filename=None):
    """Copy the contents of a file from a given URL
    to a local file and creates the destination dir if don't exists.
    """
    # check if the url it's ok
    try:
        urllib2.urlopen(url)
    except  urllib2.HTTPError, e:
        print >>sys.stderr, e, ': ', url
        sys.exit(1)
    if not os.path.exists(destdir):
        os.makedirs(destdir)
    if filename is None:
        filename = os.path.basename(urlparse.urlparse(url).path)
    dest = os.path.join(destdir, filename)
    p = ProgressSpinner()
    filename, headers = urllib.urlretrieve(url, dest, reporthook=p.next)
    p.finished()

def fetch_packages(update):
    all = config.plugins + config.packages
    for dep in all:
        tarball_filename = config.get_tarfile(dep)
        tarball_path = os.path.join(external_dir, tarball_filename)
        if dep.get('revno', None) != 'tip' and os.path.exists(tarball_path):
            print >> sys.stderr, "%s already exists" % tarball_filename
            continue
        elif dep.get('url', None):
            print "Fetching %s" % dep['name']
            sys.stdout.flush()
            download(dep['url'], external_dir, tarball_filename)
        elif dep.get('branch', None):
            tree_dir = os.path.join(external_dir, dep['name'])
            if dep['revno'] == 'tip':
                revno = []
                nice_revno = 'tip'
            else:
                revno = ['-r', dep['revno']]
                if dep['revno'][:4] == 'tag:':
                    nice_revno = dep['revno'][4:]
                else:
                    nice_revno = 'r' + dep['revno']
            cmd = ['bzr']
            if os.path.exists(tree_dir):
                if update:
                    print "Updating %s to %s" % (dep['name'], nice_revno)
                    cmd += ['pull', '-d', tree_dir]
                    cmd += revno
                    subprocess.call(cmd)
            else:
                print "Fetching %s %s" % (dep['name'], nice_revno)
                cmd += ['branch', dep['branch']]
                cmd += revno + [tree_dir]
                subprocess.call(cmd)

            if check_revno(tree_dir, dep):
                if not update and os.path.exists(tarball_path):
                    print "%s already exists" % tarball_filename
                else:
                    print "Exporting checkout to %s" % tarball_filename
                    subprocess.call(['bzr', 'export', tarball_path, tree_dir])

def fetch_pdf_docs():
    version = config.bzr['docversion']
    user_url = 'http://doc.bazaar-vcs.org/bzr.%s/downloads/pdf-en/' % version
    dev_url = 'http://doc.bazaar-vcs.org/bzr.%s/downloads/pdf-developers/' % version
    import re
    regexp = re.compile('.*href="(?P<href>.*.pdf)"')
    for url in [dev_url, user_url]:
        print "checking acces to: ", url
        f = urllib2.urlopen(url)
        hrefs = [regexp.match(line).groups()[0] for line in f.readlines() \
                    if 'href' in line and 'pdf' in line]
        f.close()
        external_dir = os.path.join(config.base, 'Documentation')
        if not os.path.exists(external_dir):
            os.makedirs(external_dir)
        for href in hrefs:
            dest = os.path.join(external_dir, doc_names[href])
            if not os.path.exists(os.path.dirname(dest)):
                os.makedirs(os.path.dirname(dest))
            p = ProgressSpinner()
            src = url+ "/" + href
            print "downloading: ", src
            filename, headers = urllib.urlretrieve(src, dest, reporthook=p.next)
            p.finished()


doc_names = {
    'bzr-en-architecture-overview.pdf':'Developer/Architecture Overview.pdf',
    'bzr-en-admin-guide.pdf':"Administrator's Guide.pdf",
    'bzr-en-developer-guide.pdf':'Developer/Developer Guide.pdf',
    'bzr-en-integration-guide.pdf':'Developer/Integration Guide.pdf',
    'bzr-en-testing-guide.pdf':'Developer/Testing Guide.pdf',
    'bzr-en-quick-reference.pdf':'Quick Reference.pdf',
    'bzr-en-tutorial-centralized.pdf': 'Centralized Workflow Tutorial.pdf',
    'bzr-en-tutorial-mini.pdf':'Bazaar in Five Minutes.pdf',
    'bzr-en-tutorial-with-launchpad.pdf':'Using Bazaar with Launchpad.pdf',
    'bzr-en-tutorial.pdf':'Tutorial.pdf',
    'bzr-en-upgrade-guide.pdf':'Upgrade Guide.pdf',
    'bzr-en-user-guide.pdf':'User Guide.pdf',
    'bzr-en-whats-new.pdf':"What's New.pdf",
    'bzr-en-user-reference.pdf':"User Reference.pdf",
}


if __name__ == '__main__':
    parser = OptionParser(usage="%prog [-c CONFIG_FILE] [-d] [-p] [-u]") 
    parser.add_option('-p', '--packages', dest='packages', action="store_true",
                      help='download dependencies')
    parser.add_option('-d', '--docs', dest='docs', action="store_true",
                      help='download documentation in pdf format')
    parser.add_option('-u', '--update', dest='update', action='store_true',
                      help='update tip branches')
    parser.add_option('-c', '--config', dest='config', action='store', default='config')
    (options, args) = parser.parse_args()

    try:
        global config
        config = __import__(options.config, globals(), locals(), [], 0)
    except ImportError:
        print "ERROR: failed to import config file"
        sys.exit(1)

    external_dir = os.path.join(config.base, 'src')

    if not options.packages and not options.docs:
        parser.print_usage()
        sys.exit(1)

    print "using %r" % external_dir
    if options.docs:
        fetch_pdf_docs()

    if options.packages:
        fetch_packages(options.update)
